﻿Imports System
Namespace KernelPerceptron
  Friend Class KernelPerceptronProgram
	Shared Sub Main(ByVal args() As String)
	  Console.WriteLine(vbLf & "Begin kernel perceptron demo " & vbLf)
	  Dim numFeatures As Integer = 2

	  Console.WriteLine("Goal is classification/prediction (-1/+1) with two predictors " & vbLf)
	  Console.WriteLine("Setting up 21 training data items " & vbLf)

	  ' donut-like data (not linearly separable)
	  Dim trainData(20)() As Double
	  trainData(0) = New Double() { 2.0, 3.0, -1 }
	  trainData(1) = New Double() { 2.0, 5.0, -1 }
	  trainData(2) = New Double() { 2.0, 7.0, -1 }
	  trainData(3) = New Double() { 3.0, 2.0, -1 }
	  trainData(4) = New Double() { 3.0, 8.0, -1 }
	  trainData(5) = New Double() { 4.0, 2.0, -1 }
	  trainData(6) = New Double() { 4.0, 8.0, -1 }
	  trainData(7) = New Double() { 5.0, 2.0, -1 }
	  trainData(8) = New Double() { 5.0, 8.0, -1 }
	  trainData(9) = New Double() { 6.0, 3.0, -1 }
	  trainData(10) = New Double() { 6.0, 5.0, -1 }
	  trainData(11) = New Double() { 6.0, 7.0, -1 }
	  trainData(12) = New Double() { 3.0, 4.0, +1 }
	  trainData(13) = New Double() { 3.0, 5.0, +1 }
	  trainData(14) = New Double() { 3.0, 6.0, +1 }
	  trainData(15) = New Double() { 4.0, 4.0, +1 }
	  trainData(16) = New Double() { 4.0, 5.0, +1 }
	  trainData(17) = New Double() { 4.0, 6.0, +1 }
	  trainData(18) = New Double() { 5.0, 4.0, +1 }
	  trainData(19) = New Double() { 5.0, 5.0, +1 }
	  trainData(20) = New Double() { 5.0, 6.0, +1 }
	  Dim numTrain As Integer = trainData.Length

	  ' test
	  Dim testData(3)() As Double
	  testData(0) = New Double() { 2.0, 4.0, -1 }
	  testData(1) = New Double() { 6.0, 6.0, -1 }
	  testData(2) = New Double() { 3.5, 4.5, +1 }
	  testData(3) = New Double() { 5.5, 5.5, +1 }
	  'testData[3] = new double[] { 5, 6, +1 };

	  Console.WriteLine("First training data is trainData[0] = (2.0, 3.0, -1) ")
	  Console.WriteLine("Last training data is trainData[20] = (5.0, 6.0, +1) " & vbLf)
	  Dim x(numFeatures - 1) As Double ' inputs
	  Dim a(numTrain - 1) As Integer ' 'wrong counters' one cell for each train data

	  Dim sigma As Double = 1.5 ' for the kernel function
	  'double sigma = 0.5;  // gives 100% accuracy on train and test

	  ' train
	  Dim kernelMatrix(numTrain - 1)() As Double ' pre-compute all Kernel
	  For i As Integer = 0 To kernelMatrix.Length - 1
		kernelMatrix(i) = New Double(numTrain - 1){}
	  Next i
	  For i As Integer = 0 To numTrain - 1
		For j As Integer = 0 To numTrain - 1
		  Dim k As Double = Kernel(trainData(i), trainData(j), sigma)
		  kernelMatrix(j)(i) = k
		  kernelMatrix(i)(j) = kernelMatrix(j)(i)
		Next j
	  Next i

	  Console.WriteLine("Starting kernel perceptron training, sigma = 1.5 ")
	  Dim maxEpoch As Integer = 10
	  Dim epoch As Integer = 0
	  Do While epoch < maxEpoch
		For i As Integer = 0 To numTrain - 1 ' each training data
		  Dim di As Integer = CInt(Fix(trainData(i)(numFeatures))) ' get desired/correct class -1 or +1

		  Dim y As Double = 0.0 ' weighted sum of kernel results
		  For j As Integer = 0 To numTrain - 1 ' each other train items
			Dim dj As Integer = CInt(Fix(trainData(j)(numFeatures)))
			Dim kern As Double = kernelMatrix(i)(j)
			y += a(j) * dj * kern
		  Next j

		  ' update curr wrong-counters if predicted y is wrong
		  If (di = -1 AndAlso y >= 0.0) OrElse (di = 1 AndAlso y <= 0.0) Then
			a(i) += 1 ' increment wrong counter for this train data
		  End If
		Next i ' i, each train data

		epoch += 1
	  Loop ' while
	  ' end training

	  Console.WriteLine("Training complete " & vbLf)

	  Console.WriteLine("First two wrong-counters: ")
	  Console.WriteLine("a[0] = " & a(0) & "   a[1] = " & a(1))
	  Console.WriteLine("Last two wrong-counters: ")
	  Console.WriteLine("a[19] = " & a(19) & "  a[20] = " & a(20))

	  Dim trainAcc As Double = Accuracy(trainData, a, trainData, sigma, False) ' silent
	  Console.WriteLine(vbLf & "Accuracy on training data = " & trainAcc.ToString("F4"))

	  Console.WriteLine(vbLf & "Analyzing test data: ")
	  Dim testAcc As Double = Accuracy(testData, a, trainData, sigma, True) ' verbose
	  'Console.WriteLine("Accuracy on test data     = " + testAcc.ToString("F4"));

	  Console.WriteLine(vbLf & "End kernel perceptron demo " & vbLf)
	  Console.ReadLine()

	End Sub ' Main

	Private Shared Function Kernel(ByVal d1() As Double, ByVal d2() As Double, ByVal sigma As Double) As Double
	  ' RBF kernel. assumes class label in last cell
	  Dim num As Double = 0.0
	  For i As Integer = 0 To d1.Length - 2 ' note -1
		num += (d1(i) - d2(i)) * (d1(i) - d2(i))
	  Next i
	  Dim denom As Double = 2.0 * sigma * sigma
	  Dim z As Double = num / denom
	  Dim result As Double = Math.Exp(-z)
	  Return result
	End Function

	Private Shared Function Accuracy(ByVal data()() As Double, ByVal a() As Integer, ByVal trainData()() As Double, ByVal sigma As Double, ByVal verbose As Boolean) As Double
	  ' assumes data is like 2.0, 3.5  +1 (actual class in last cell)
	  Dim numFeatures As Integer = data(0).Length - 1
	  Dim numCorrect As Integer = 0
	  Dim numWrong As Integer = 0

	  For i As Integer = 0 To data.Length - 1 ' i is data to evaluate
		Dim d As Integer = CInt(Fix(data(i)(numFeatures))) ' get desired/correct class -1 or +1

		Dim y As Double = 0.0
		For j As Integer = 0 To trainData.Length - 1 ' use all train items
		  Dim d_other As Integer = CInt(Fix(trainData(j)(numFeatures))) ' +1 or -1
		  y += a(j) * d_other * Kernel(data(i), trainData(j), sigma)
		Next j

		If (d = -1 AndAlso y >= 0.0) OrElse (d = 1 AndAlso y <= 0.0) Then
		  numWrong += 1
		Else
		  numCorrect += 1
		End If

		If verbose = True Then
		  Console.Write("Input: ")
		  For j As Integer = 0 To data(i).Length - 2
			Console.Write(data(i)(j).ToString("F1") & " ")
		  Next j
		  If data(i)(data(i).Length - 1) = 1 Then
			Console.Write("+")
		  End If
		  Console.Write(data(i)(data(i).Length - 1).ToString("F0"))

		  Console.Write("   y calc = ")
		  If y >= 0.0 Then
			  Console.Write("+")
		  End If
		  Console.Write(y.ToString("F4"))
		  Console.Write("  predicted = ")
		  If y < 0.0 Then
			  Console.Write("-1")
		  Else
			  Console.Write("+1")
		  End If

		  If (d = -1 AndAlso y >= 0.0) OrElse (d = 1 AndAlso y <= 0.0) Then
			Console.WriteLine("  wrong")
		  Else
			Console.WriteLine("  correct")
		  End If
		End If

	  Next i ' for each data
	  Return (1.0 * numCorrect) / (numCorrect + numWrong)
	End Function

  End Class ' Program

End Namespace ' ns
